﻿using System.Collections;

var list = new List<int>() { 1, 2, 3, 4, 5 };
Console.WriteLine($"<< {list[2]} <<");
Console.WriteLine($"<< {list[^1]} <<");
Console.WriteLine();

var simpleList = new SimpleList();

Console.WriteLine($"<< {simpleList[2]} <<");
Console.WriteLine($"<< {simpleList[^1]} <<");
Console.WriteLine();

var indexList = new IndexList();

Console.WriteLine($"<< {indexList[2]} <<");
Console.WriteLine($"<< {indexList[^1]} <<");
Console.WriteLine();

// Error: Argument 1: cannot convert from 'System.Range' to 'int'
// Console.WriteLine($"<< {simpleList[1..^2]} <<");
// Error: Argument 1: cannot convert from 'System.Range' to 'System.Index'
// Console.WriteLine($"<< {indexList[1..^2]} <<");

var sliceList = new SliceList();

Console.WriteLine($"<< {sliceList[1..^2]} <<");
Console.WriteLine();

var rangeList = new RangeSliceList();
Console.WriteLine($"<< {rangeList[1..^2]} <<");
Console.WriteLine();

// Console.WriteLine($"<< {list[1..^2]} <<");

Console.WriteLine("---");
Console.WriteLine($"rangeList[1..^2]:"); _ = rangeList[1..^2];
Console.WriteLine($"rangeList[1..2]:"); _ = rangeList[1..2];
Console.WriteLine($"rangeList[^2..^0]:"); _ = rangeList[^2..^0];
Console.WriteLine($"rangeList[1..]:"); _ = rangeList[1..];
Console.WriteLine($"rangeList[..2]:"); _ = rangeList[..2];
Console.WriteLine($"rangeList[..^2]:"); _ = rangeList[..^2];
Console.WriteLine($"rangeList[..]:"); _ = rangeList[..];

class SimpleList : IReadOnlyList<string> {
	protected List<string> _list = new() { "one", "two", "three", "four", "five" };

	// IReadOnlyList<string>:

	public string this[int index] {
		get {
			Console.WriteLine($"SimpleList.this[int index == {index}]");
			return _list[index];
		}
	}

	public int Count {
		get {
			Console.WriteLine($"SimpleList.get_Count -> {_list.Count}");
			return _list.Count;
		}
	}

	public IEnumerator<string> GetEnumerator() => _list.GetEnumerator();

	IEnumerator IEnumerable.GetEnumerator() => GetEnumerator();
}

class IndexList : SimpleList {
	public string this[Index index] {
		get {
			Console.WriteLine($"IndexList.this[Index index == {index} ~ {{ Value = {index.Value}, IsFromEnd = {index.IsFromEnd}}})]");
			return _list[index];
		}
	}
}

class SliceList : IndexList {
	public IReadOnlyList<string> Slice(int index, int length) {
		Console.WriteLine($"SliceList.Slice(int index == {index}, int length == {length})");
		return _list.GetRange(index, length);
	}
}

class RangeSliceList : SliceList {
	public IReadOnlyList<string> this[Range range] {
		get {
			Console.WriteLine($"RangeSliceList.this[Range range == {range} ~ {{ Start = {range.Start}, End = {range.End}}}]");
			var (offset, count) = range.GetOffsetAndLength(_list.Count);
			Console.WriteLine($"  -> offset: {offset}, count: {count}");
			return _list.GetRange(offset, count);
		}
	}
}

//static class ListExtensions {
//	public static IReadOnlyList<T> Slice<T>(this List<T> list, int index, int length) {
//		Console.WriteLine($"ListExtensions.Slice<{typeof(T)}>(int index == {index}, int length == {length})");
//		return list.GetRange(index, length);
//	}
//}
